SUBMERCADOS = ["SUDESTE", "SUL", "NORDESTE", "NORTE"]
REES = ["SUDESTE",
        "SUL",
        "NORDESTE",
        "NORTE",
        "ITAIPU",
        "MADEIRA",
        "TPIRES",
        "BMONTE",
        "MAN-AP",
        "PARANA",
        "IGUACU",
        "PRNPANEMA"]
NUM_CENARIOS = 2000
MAX_ANOS_ESTUDO = 30
MAX_ITERS = 200
NUM_PATAMARES = 3
NUM_VARIAVEIS_MEDIAS = 52
NUM_VARIAVEIS_CUSTO_PMO = 16
ORDEM_MAX_PARP = 6
MESES = ["JANEIRO",
         "FEVEREIRO",
         "MARCO",
         "ABRIL",
         "MAIO",
         "JUNHO",
         "JULHO",
         "AGOSTO",
         "SETEMBRO",
         "OUTUBRO",
         "NOVEMBRO",
         "DEZEMBRO"]
